package ru.dbotthepony.mc.otm.client.render

import com.mojang.blaze3d.systems.RenderSystem
import com.mojang.blaze3d.vertex.BufferBuilder
import com.mojang.blaze3d.vertex.BufferUploader
import com.mojang.blaze3d.vertex.ByteBufferBuilder
import com.mojang.blaze3d.vertex.DefaultVertexFormat
import com.mojang.blaze3d.vertex.PoseStack
import com.mojang.blaze3d.vertex.VertexFormat
import it.unimi.dsi.fastutil.floats.Float2ObjectMap
import net.minecraft.client.gui.Font
import net.minecraft.client.renderer.GameRenderer
import net.minecraft.network.chat.Component
import org.lwjgl.opengl.GL11.GL_LESS
import ru.dbotthepony.kommons.math.RGBAColor
import kotlin.math.PI
import kotlin.math.acos
import kotlin.math.cos
import kotlin.math.pow
import kotlin.math.sign
import kotlin.math.sin

private const val LINE_WIDTH = 1f
private const val HIGHLIGHT_WIDTH = 3f

private val TEXT_BACKGROUND = RGBAColor(0f, 0f, 0f, 0.5f)
private val LINE_COLOR = RGBAColor(1f, 0f, 1f, 0.25f)
private val LINE_TEXT_COLOR = RGBAColor(1f, 0f, 1f, 0.5f)

data class ChartMouseLabels(
	val mouseX: Float,
	val mouseY: Float,
	val labels: (Int) -> List<Component>,
	val font: Font,
	val color: RGBAColor = RGBAColor.WHITE,
	val pillarColor: RGBAColor = RGBAColor.WHITE,
	val textColor: RGBAColor = RGBAColor.WHITE,
	val textBackgroundColor: RGBAColor = TEXT_BACKGROUND
)

data class ChartLevelLabels(
	val labels: Map<Float, Component>,
	val font: Font,
	val lineColor: RGBAColor = RGBAColor.DARK_GREEN,
	val textColor: RGBAColor = RGBAColor.DARK_GREEN,
	val textGravity: RenderGravity = RenderGravity.TOP_RIGHT,
	val textOutline: RGBAColor = RGBAColor.BLACK
)

private val BYTE_BUFFER_BUILDER = ByteBufferBuilder(786_432)

fun renderChart(
	poseStack: PoseStack,
	normalized: FloatArray,
	width: Float,
	height: Float,
	color: RGBAColor = RGBAColor.WHITE,
	labels: ChartMouseLabels? = null,
	levelLabels: ChartLevelLabels? = null,
	x: Float = 0f,
	y: Float = 0f,
) {
	val builder = BufferBuilder(BYTE_BUFFER_BUILDER, VertexFormat.Mode.QUADS, DefaultVertexFormat.POSITION_COLOR)
	val step = width / normalized.size

	val pose = poseStack.last().pose
	val checkLabels = labels != null && labels.mouseY in 0f .. height && labels.mouseX in 0f .. width

	var drawLabel = -1
	var drawPointX = 0f
	var drawPointY = 0f

	if (levelLabels != null) {
		for ((level, label) in levelLabels.labels) {
			val y0 = y + (1f - level) * height

			builder.vertex(pose, x + width, LINE_WIDTH / 2f + y0, 0f).color(levelLabels.lineColor)
			builder.vertex(pose, x + width, -LINE_WIDTH / 2f + y0, 0f).color(levelLabels.lineColor)
			builder.vertex(pose, x, -LINE_WIDTH / 2f + y0, 0f).color(levelLabels.lineColor)
			builder.vertex(pose, x, LINE_WIDTH / 2f + y0, 0f).color(levelLabels.lineColor)
		}
	}

	for (i in 0 until normalized.size - 1) {
		val y0 = y + (1f - normalized[i]) * height
		val y1 = y + (1f - normalized[i + 1]) * height

		val x0 = x + width - (i + 0.5f) * step
		val x1 = x + width - (i + 1.5f) * step

		val xDiff = x1 - x0
		val yDiff = y1 - y0

		if (yDiff.sign == 0f) {
			builder.vertex(pose, x0, y0 + LINE_WIDTH / 2f, 0f).color(color)
			builder.vertex(pose, x0, y0 - LINE_WIDTH / 2f, 0f).color(color)
			builder.vertex(pose, x1, y1 - LINE_WIDTH / 2f, 0f).color(color)
			builder.vertex(pose, x1, y1 + LINE_WIDTH / 2f, 0f).color(color)
		} else {
			val length = (xDiff.pow(2f) + yDiff.pow(2f)).pow(0.5f)
			val angle = acos(xDiff / length) * yDiff.sign - PI / 2f // rotate 90 deg

			val xDisp = cos(angle).toFloat() * LINE_WIDTH / 2f
			val yDisp = sin(angle).toFloat() * LINE_WIDTH / 2f

			builder.vertex(pose, x0 + xDisp, y0 + yDisp, 0f).color(color)
			builder.vertex(pose, x0 - xDisp, y0 - yDisp, 0f).color(color)
			builder.vertex(pose, x1 - xDisp, y1 - yDisp, 0f).color(color)
			builder.vertex(pose, x1 + xDisp, y1 + yDisp, 0f).color(color)
		}

		//graphics.renderRect(x0, y0, LINE_WIDTH, LINE_WIDTH)

		if (checkLabels && drawLabel == -1 && labels!!.mouseX in x1 .. x0) {
			drawLabel = i
			drawPointX = x0 // linearInterpolation(0.5f, x0, x1)
			drawPointY = y0 // linearInterpolation(0.5f, y0, y1)
		}
	}

	RenderSystem.setShader(GameRenderer::getPositionColorShader)
	// RenderSystem.setShaderColor(color.red, color.green, color.blue, color.alpha)
	RenderSystem.disableDepthTest()
	RenderSystem.disableBlend()
	RenderSystem.setShaderColor(1f, 1f, 1f, 1f)

	BufferUploader.drawWithShader(builder.buildOrThrow())

	RenderSystem.enableDepthTest()
	RenderSystem.depthFunc(GL_LESS)

	if (levelLabels != null && levelLabels.labels.isNotEmpty()) {
		for ((level, label) in levelLabels.labels) {
			val y0 = y + (1f - level) * height

			val tX = x + levelLabels.textGravity.repositionX(width, levelLabels.font.width(label).toFloat()) - 1f
			val tY = y0 - LINE_WIDTH - 1f - levelLabels.font.lineHeight + levelLabels.textGravity.repositionY(levelLabels.font.lineHeight * 2f + LINE_WIDTH + 2f, levelLabels.font.lineHeight.toFloat()) + 1f

			levelLabels.font.draw(
				poseStack,
				label,
				x = tX,
				y = tY,
				color = levelLabels.textColor,
				drawOutline = true,
				outlineZ = -1f,
				outlineColor = levelLabels.textOutline
			)
		}
	}

	if (drawLabel != -1) {
		renderRect(pose, drawPointX - LINE_WIDTH / 2f, y, LINE_WIDTH, height, color = labels!!.pillarColor)
		renderRect(pose, drawPointX - HIGHLIGHT_WIDTH / 2f, drawPointY - HIGHLIGHT_WIDTH / 2f, HIGHLIGHT_WIDTH, HIGHLIGHT_WIDTH, color = labels.color)

		val label = labels.labels(drawLabel)

		if (label.isNotEmpty()) {
			val fWidth = label.maxOf { labels.font.width(it).toFloat() + 2f }
			val fHeight = labels.font.lineHeight.toFloat() * label.size + 2f

			val anchorX = labels.mouseX // drawPointX
			var anchorY = labels.mouseY - fHeight - 2f

			renderRect(pose, anchorX - fWidth / 2f, anchorY - 2f, fWidth, fHeight + 4f, color = labels.textBackgroundColor)

			label.forEach {
				labels.font.draw(poseStack, it, anchorX, anchorY, gravity = RenderGravity.TOP_CENTER, color = labels.textColor)
				anchorY += labels.font.lineHeight
			}
		}
	}
}
